#define _GNU_SOURCE
#include <openssl/ssl.h>
#include <openssl/err.h>

#include <signal.h>
#include <unistd.h>

#include <sys/socket.h>
#include <sys/mount.h>
#include <netinet/ip.h>
#include <fcntl.h>
#include <stdio.h>

#include <sched.h>
#include <dirent.h>

#include "url.h"

int shut = -1;

const char HEADER_OKAY_GEMINI[] = "20 text/gemini\r\n\0";
const char HEADER_ERR_GEMINI[] = "50 ERR\r\n\0";


void
hand_sig(int signal)
{
	printf("Closing...\n");
	shut = 1;
}


int
create_sock(int port)
{
	int sock, err;
	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = htonl(INADDR_ANY);

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (!sock){
		printf("error on sock");
		exit(EXIT_FAILURE);
	}

	err = bind(sock, (struct sockaddr*)&addr, sizeof(addr));
	if (err < 0){ 
		printf("error on bind %i", err);
		exit(EXIT_FAILURE);
	}
	err = listen(sock, 10);
	if (err < 0){
		printf("error on listen %i", err);
		exit(EXIT_FAILURE);
	}
	return sock;
}


int
setup_namespace()
{
	mkdir(".tmp-data", 0760);
	mkdir("data/old_dir", 0760);
	if(unshare(CLONE_NEWNS|CLONE_NEWUSER)<0){
		printf("FAILED UNSHARE: %s \n", strerror(errno));
		return -1;
	}
	if(mount("data", ".tmp-data", NULL, MS_BIND | MS_PRIVATE, NULL)<0){
		printf("FAILED BIND MOUNT: %s \n", strerror(errno));
		return -2;
	}
	if(chdir(".tmp-data")<0){
		printf("FAILED CHDIR .tmp-data: %s \n", strerror(errno));
		return -3;
	}
	if(pivot_root(".", "old_dir")<0){
		printf("FAILED PIVOT_ROOT: %s \n", strerror(errno));
		return -4;
	}
	if(umount2("old_dir", MNT_DETACH)<0){
		printf("FAILED OLD_DIR UMOUNT: %s \n", strerror(errno));
		return -5;
	}
	chdir("/");
	rmdir("old_dir");
	printf("%s\n", strerror(errno));
	printf(get_current_dir_name());

	return 0;
}


int
configure_ctx(SSL_CTX *ctx)
{
	int err = 0;

	SSL_CTX_set_min_proto_version(ctx, TLS1_3_VERSION);
	SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, NULL);
	SSL_CTX_set_options(ctx, SSL_OP_NO_RENEGOTIATION);

	//SSL_CTX_set_ciphersuites(ctx, "");
	err = SSL_CTX_use_PrivateKey_file(ctx, "key.pem", SSL_FILETYPE_PEM);
	if(err < 0){
		printf("Pepperoni pizza!\n");
		ERR_print_errors_fp(stdout);
		return -10;
	}
	err = SSL_CTX_use_certificate_file(ctx, "cert.pem", SSL_FILETYPE_PEM);
	if(err < 0){
		printf("Certificati anulati\n");
		ERR_print_errors_fp(stdout);
		return -11;
	}
	err = SSL_CTX_check_private_key(ctx);
	if(err < 0){
		printf("The combinati e invalidate\n");
		ERR_print_errors_fp(stdout);
		return -12;
	}
}


void
cleanup(SSL_CTX *ctx,int sock)
{
	close(sock);
	SSL_CTX_free(ctx);
	printf("Closed server\n");
}


int
main()
{
	int sock = 0;
	int err = 0;

	signal(SIGINT, hand_sig);

	SSL_CTX *ctx = SSL_CTX_new(TLS_server_method());
	configure_ctx(ctx);

	sock = create_sock(1965);
	if(setup_namespace()<0){
		cleanup(ctx, sock);
		return -50;
	}

	ERR_print_errors_fp(stdout);

	while(shut < 0){
		srv_run(ctx, sock);
	}

	cleanup(ctx, sock);
	return 0;
}


int
serve_gmni(char *path, SSL *ssl){
	if(is_index(path)) return serve_gmni("/index.gmi", ssl);

	printf("Serving '%s' \n", path);

	int gmni_fd = open(path, O_RDONLY);
	if(gmni_fd < 0){
		printf("FILE: '%s' %i NOT FOUND\n", path, gmni_fd);
		SSL_write(ssl, HEADER_ERR_GEMINI, strlen(HEADER_ERR_GEMINI));
		return -1;
	}
	SSL_write(ssl, HEADER_OKAY_GEMINI, strlen(HEADER_OKAY_GEMINI));
	char buf[4096];
	int buflen = read(gmni_fd, buf, 4096);
	while (buflen>0){
		SSL_write(ssl, buf, buflen);
		buflen = read(gmni_fd, buf, 4096);
	}
	return 0;
}

void
srv_run(SSL_CTX *ctx, int sock){
	struct sockaddr_in addr;
	struct url parsed_url;
	int err = 0;
	uint len = sizeof(addr);
	char request[1025];
	SSL *ssl;

	int client = accept(sock, (struct sockaddr *)&addr, &len);
	//Podria deberia agregar el cliente a un 
	//Aca podria forkear ya
	ssl = SSL_new(ctx);
	SSL_set_fd(ssl, client);
	SSL_accept(ssl);

	SSL_read(ssl, request, 1024);
	request[1024] = '\0';	//obligatory null terminator
	err = url_parse(request, &parsed_url);
	if(err){
		printf("[err] malformed request: %s %d\n", request, err);
	}else{
		serve_gmni(parsed_url.path, ssl);
		printf("Serving: \n");
		printf("scheme: %s %i\n", parsed_url.scheme, parsed_url.scheme_len);
		printf("host: %s %i\n", parsed_url.host, parsed_url.host_len);
		printf("path: %s %i\n", parsed_url.path, parsed_url.path_len);
		printf("-----\n");
	}

	ERR_print_errors_fp(stdout);
	//De esto se deberia encargar el child idealmente...
	SSL_shutdown(ssl);
	SSL_free(ssl);
	close(client);
}
